// Copyright 2017 The Prometheus Authors
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package uyuni

import (
	"context"
	"fmt"
	"net/http"
	"time"

	"github.com/go-kit/kit/log"
	"github.com/go-kit/kit/log/level"
	"github.com/kolo/xmlrpc"
	"github.com/pkg/errors"
	"github.com/prometheus/common/model"

	"github.com/prometheus/prometheus/discovery/refresh"
	"github.com/prometheus/prometheus/discovery/targetgroup"
)

const (
	uyuniLabel             = model.MetaLabelPrefix + "uyuni_"
	uyuniLabelEntitlements = uyuniLabel + "entitlements"
)

// DefaultSDConfig is the default Uyuni SD configuration.
var DefaultSDConfig = SDConfig{
	RefreshInterval: model.Duration(1 * time.Minute),
}

// SDConfig is the configuration for Uyuni based service discovery.
type SDConfig struct {
	Host            string         `yaml:"host"`
	User            string         `yaml:"username"`
	Pass            string         `yaml:"password"`
	RefreshInterval model.Duration `yaml:"refresh_interval,omitempty"`
}

// Uyuni API Response structures
type clientRef struct {
	Id   int    `xmlrpc:"id"`
	Name string `xmlrpc:"name"`
}

type clientDetail struct {
	Id           int      `xmlrpc:"id"`
	Hostname     string   `xmlrpc:"hostname"`
	Entitlements []string `xmlrpc:"addon_entitlements"`
}

type exporterConfig struct {
	Enabled bool `xmlrpc:"enabled"`
}

type formulaData struct {
	NodeExporter     exporterConfig `xmlrpc:"node_exporter"`
	PostgresExporter exporterConfig `xmlrpc:"postgres_exporter"`
}

// UnmarshalYAML implements the yaml.Unmarshaler interface.
func (c *SDConfig) UnmarshalYAML(unmarshal func(interface{}) error) error {
	*c = DefaultSDConfig
	type plain SDConfig
	err := unmarshal((*plain)(c))

	if err != nil {
		return err
	}
	if c.Host == "" {
		return errors.New("Uyuni SD configuration requires a Host")
	}
	if c.User == "" {
		return errors.New("Uyuni SD configuration requires a Username")
	}
	if c.Pass == "" {
		return errors.New("Uyuni SD configuration requires a Password")
	}
	if c.RefreshInterval <= 0 {
		return errors.New("Uyuni SD configuration requires RefreshInterval to be a positive integer")
	}
	return nil
}

// Attempt to login in SUSE Manager Server and get an auth token
func Login(rpcclient *xmlrpc.Client, user string, pass string) (string, error) {
	var result string
	err := rpcclient.Call("auth.login", []interface{}{user, pass}, &result)
	return result, err
}

// Logout from SUSE Manager API
func Logout(rpcclient *xmlrpc.Client, token string) error {
	err := rpcclient.Call("auth.logout", token, nil)
	return err
}

// Get client list
func ListSystems(rpcclient *xmlrpc.Client, token string) ([]clientRef, error) {
	var result []clientRef
	err := rpcclient.Call("system.listSystems", token, &result)
	return result, err
}

// Get client details
func GetSystemDetails(rpcclient *xmlrpc.Client, token string, systemId int) (clientDetail, error) {
	var result clientDetail
	err := rpcclient.Call("system.getDetails", []interface{}{token, systemId}, &result)
	return result, err
}

// List client FQDNs
func ListSystemFQDNs(rpcclient *xmlrpc.Client, token string, systemId int) ([]string, error) {
	var result []string
	err := rpcclient.Call("system.listFqdns", []interface{}{token, systemId}, &result)
	return result, err
}

// Get formula data for a given system
func getSystemFormulaData(rpcclient *xmlrpc.Client, token string, systemId int, formulaName string) (formulaData, error) {
	var result formulaData
	err := rpcclient.Call("formula.getSystemFormulaData", []interface{}{token, systemId, formulaName}, &result)
	return result, err
}

// Discovery periodically performs Uyuni API requests. It implements
// the Discoverer interface.
type Discovery struct {
	*refresh.Discovery
	client     *http.Client
	interval   time.Duration
	sdConfig   *SDConfig
	logger     log.Logger
}

// NewDiscovery returns a new file discovery for the given paths.
func NewDiscovery(conf *SDConfig, logger log.Logger) *Discovery {
	d := &Discovery{
		interval: time.Duration(conf.RefreshInterval),
		sdConfig: conf,
		logger: logger,
	}
	d.Discovery = refresh.NewDiscovery(
		logger,
		"uyuni",
		time.Duration(conf.RefreshInterval),
		d.refresh,
	)
	return d
}

func (d *Discovery) refresh(ctx context.Context) ([]*targetgroup.Group, error) {

	config := d.sdConfig
	apiUrl := config.Host + "/rpc/api"

	rpcclient, _ := xmlrpc.NewClient(apiUrl, nil)

	token, err := Login(rpcclient, config.User, config.Pass)
	if err != nil {
		return nil, errors.Wrap(err, "Unable to login to SUSE Manager API")
	}

	clientList, err := ListSystems(rpcclient, token)
	if err != nil {
		return nil, errors.Wrap(err, "Unable to get list of systems")
	}

	tg := &targetgroup.Group{
		Source: config.Host,
	}

	if len(clientList) == 0 {
		fmt.Printf("\tFound 0 systems.\n")
	} else {
		for _, client := range clientList {
			fqdns := []string{}
			formulas := formulaData{}
			details, err := GetSystemDetails(rpcclient, token, client.Id)
			if err != nil {
				level.Error(d.logger).Log("msg", "Unable to get system details","clientId", client.Id, "err", err)
				continue;
			}
			// Check if system is to be monitored
			for _, v := range details.Entitlements {
				if v == "monitoring_entitled" {
					fqdns, err = ListSystemFQDNs(rpcclient, token, client.Id)
					formulas, err = getSystemFormulaData(rpcclient, token, client.Id, "prometheus-exporters")
					if (formulas.NodeExporter.Enabled) {
						labels := model.LabelSet{}
						addr := fmt.Sprintf("%s:%d", fqdns[len(fqdns)-1], 9100)
						labels[model.AddressLabel] = model.LabelValue(addr)
						tg.Targets = append(tg.Targets, labels)
					}
					if (formulas.PostgresExporter.Enabled) {
						labels := model.LabelSet{}
						addr := fmt.Sprintf("%s:%d", fqdns[len(fqdns)-1], 9187)
						labels[model.AddressLabel] = model.LabelValue(addr)
						tg.Targets = append(tg.Targets, labels)
					}
				}
			}

			level.Debug(d.logger).Log("msg", "Found system", "host", details.Hostname, "entitlements", fmt.Sprintf("%+v", details.Entitlements), "FQDN", fmt.Sprintf("%+v", fqdns), "formulas", fmt.Sprintf("%+v", formulas))
		}
	}
	Logout(rpcclient, token)
	return []*targetgroup.Group{tg}, nil
}
